@@ -2,7 +2,7 @@
 #
 #  File::NFSLock - bdpO - NFS compatible (safe) locking utility
 #  
-#  $Id: NFSLock.pm,v 1.2 2001/05/25 05:53:37 rhandom Exp $
+#  $Id: NFSLock.pm,v 1.14 2001/07/31 18:23:17 pauls Exp $
 #  
 #  Copyright (C) 2001, Paul T Seamons
 #                      paul@seamons.com
@@ -26,46 +26,83 @@ package File::NFSLock;
 
 use strict;
 use Exporter ();
-use vars qw(@ISA @EXPORT_OK $VERSION);
+use vars qw(@ISA @EXPORT_OK $VERSION $TYPES $extended $LOCK_EXTENSION $errstr);
 
 @ISA = qw(Exporter);
 @EXPORT_OK = qw(uncache);
 
-$VERSION = '1.00';
+$VERSION = '1.10';
+
+### hash of types
+$TYPES = {
+  BLOCKING    => 'BL',
+  BL          => 'BL',
+  EXCLUSIVE   => 'BL',
+  EX          => 'BL',
+  NONBLOCKING => 'NB',
+  NB          => 'NB',
+  SHARED      => 'SH',
+  SH          => 'SH',
+};
+$extended = undef; # are Fcntl constants loaded ?
+$LOCK_EXTENSION = '.NFSLock'; # customizable extension
 
 ###----------------------------------------------------------------###
 
-sub new ($$) {
+sub new {
+  $errstr = undef;
+
   my $type  = shift;
   my $class = ref($type) || $type || __PACKAGE__;
   my $self  = {};
-  $self->{file}      = shift;
-  $self->{lock_file} = $self->{file} . '.NFSLock';
-  $self->{lock_type} = shift;
-  $self->{blocking_timeout} = shift || 0;
-  
-  die "Usage: my \$f = O::Lock->new('/pathtofile/file',
-        'BLOCKING|BL|NONBLOCKING|NB|SHARED|SH', [timeout]);"
-    unless length($self->{file})
-      && $self->{lock_type} =~ /^((NON)?BLOCKING|NB|BL|SH(ARED)?)$/;
 
-  die "Shared Locks are not yet implemented (but soon)"
-    if $self->{lock_type} =~ /^S/;
+  ### allow for arguments by hash ref or serially
+  if( @_ && ref $_[0] ){
+    $self = shift;
+  }else{
+    $self->{file}      = shift;
+    $self->{lock_type} = shift;
+    $self->{blocking_timeout}   = @_ ? shift : undef;
+    $self->{stale_lock_timeout} = @_ ? shift : undef;
+  }
+  $self->{blocking_timeout}   ||= 0;
+  $self->{stale_lock_timeout} ||= 0;
+
+  ### if passed a numerical lock type, load contants
+  if( $self->{lock_type} =~ /^\d+/ && ! $extended ){
+    $extended = 1;
+    require "Fcntl.pm";
+    $TYPES->{ Fcntl::LOCK_SH() } = 'SH';
+    $TYPES->{ Fcntl::LOCK_EX() } = 'BL';
+    $TYPES->{ Fcntl::LOCK_NB() } = 'NB';
+  }
+
+  ### quick usage check
+  die $errstr = "Usage: my \$f = File::NFSLock->new('/pathtofile/file',\n"
+    ."'BLOCKING|EXCLUSIVE|NONBLOCKING|SHARED', [blocking_timeout, stale_lock_timeout]);\n"
+      ."(You passed \"$self->{file}\" and \"$self->{lock_type}\")"
+        unless length($self->{file}) && exists $TYPES->{ $self->{lock_type} };
+  $self->{lock_type} = $TYPES->{ $self->{lock_type} };
+
+  die $errstr = "Shared Locks are not yet implemented (but soon)"
+    if $self->{lock_type} eq 'SH';
   
+  ### set some utility files
   $self->{local_file} = local_file( $self->{file} );
+  $self->{lock_file}  = $self->{file} . $LOCK_EXTENSION;
 
   ### open a local temporary file
   open_local_file( $self->{local_file} )
     or return undef;
 
-  ### non blocking lock (return undef on failure)
-  if( $self->{lock_type} =~ /^N/ ){
-    do_lock( $self->{lock_file}, $self->{local_file} )
-      or return undef;
+  ### nonblocking lock (return undef on failure)
+  if( $self->{lock_type} eq 'NB' ){
+    do_lock( $self->{lock_file}, $self->{local_file}, $self->{stale_lock_timeout} )
+      or do { $errstr = "Couldn't get a lock on a NONBLOCKING lock"; return undef; };
 
   ### blocking lock, wait until it's done
   }else{
-    do_lock_blocking( $self->{lock_file}, $self->{local_file}, $self->{blocking_timeout} )
+    do_lock_blocking( $self->{lock_file}, $self->{local_file}, $self->{blocking_timeout}, $self->{stale_lock_timeout} )
       or return undef;
 
   }
@@ -100,41 +137,63 @@ sub local_file ($) {
 }
 
 sub open_local_file ($) {
+  $errstr = undef;
   my $local_file = shift;
-  open (_LOCK,">>$local_file") or return undef;
-  print _LOCK "Pid [$$]\nHost [$ENV{HOSTNAME}]\nFile [$local_file]\n";
+  open (_LOCK,">>$local_file") or do { $errstr = "Couldn't open \"$local_file\" [$!]"; return undef; };
+  print _LOCK "Pid [$$]\nHost [$ENV{HOSTNAME}]\nFile [$local_file]\n"; # trace information
   close(_LOCK);
   return 1;
 }
 
-sub do_lock ($$) {
-  my $lock_file  = shift;
-  my $local_file = shift;
+sub do_lock {
+  $errstr = undef;
+  my $lock_file     = shift;
+  my $local_file    = shift;
+  my $stale_timeout = shift || 0;
+  my $recurse       = shift || 0;
 
   ### try a hard link
   link( $local_file, $lock_file);
 
   ### did it work (two files pointing to local_file)
-  my $success = ( (stat($local_file))[3] == 2 );
+  my $success = ( -e $local_file && (stat($local_file))[3] == 2 );
   unlink $local_file;
 
+  ### remove an old lockfile if it is older than the stale_timeout
+  if( ! $success && $stale_timeout > 0 ){
+    if( !-e $lock_file || time() - (stat($lock_file))[9] > $stale_timeout ){ # check mtime
+      if( ! unlink($lock_file) ){
+        $errstr = "Can't unlink stale lock file \"$lock_file\" [$!]";
+        return undef;
+      }elsif( ++$recurse >= 10 ){
+        $errstr = "Max retries reached trying to remove stale lockfile";
+        return undef;
+      }else{
+        return do_lock( $lock_file, $local_file, $stale_timeout,$recurse );
+      }
+    }
+  }
+
   return $success;
 }
 
 sub do_lock_blocking {
-  my $lock_file  = shift;
-  my $local_file = shift;
-  my $timeout    = shift;
+  $errstr = undef;
+  my $lock_file     = shift;
+  my $local_file    = shift;
+  my $timeout       = shift;
+  my $stale_timeout = shift || 0;
   my $start_time = $timeout ? time() : 0;
 
-  while( ! do_lock( $lock_file, $local_file ) ){
+  while( ! do_lock( $lock_file, $local_file, $stale_timeout ) ){
 
     ### wait a moment
     sleep(1);
 
     ### but don't wait past the time out
     if( $timeout && (time() - $start_time) > $timeout ){
-      return 0;
+      $errstr = "Timed out waiting for blocking lock";
+      return undef;
     }
 
     ### reopen the file
@@ -172,7 +231,16 @@ File::NFSLock - perl module to do NFS (or not) locking
   my $file = "somefile";
 
   ### set up a lock - lasts until object looses scope
-  if( defined(my $lock = File::NFSLock->new($file,"NONBLOCKING")) ){
+  if( defined(my $lock = File::NFSLock->new({
+    file      => $file,
+    lock_type => "NONBLOCKING"
+    blocking_timeout   => 10,      # 10 sec
+    stale_lock_timeout => 30 * 60, # 60 min
+    })) ){
+    
+    ### OR
+    ### my $lock = File::NFSLock->new($file,"NONBLOCKING",10,30*60)
+
     
     ### do write protected stuff on $file
     ### at this point $file is uncached from NFS (most recent)
@@ -191,6 +259,8 @@ File::NFSLock - perl module to do NFS (or not) locking
     ### OR
     ### undef $lock;
     ### OR let $lock go out of scope
+  }else{
+    die "I couldn't lock the file [$File::NFSLock::errstr]";
   }
 
 
@@ -215,26 +285,49 @@ place until the lock object goes out of scope (or calls the unlock
 method).
 
 A lock object is created by calling the new method and passing two 
-or three parameters:
+to four parameters in the following manner:
+
+  my $lock = File::NFSLock->new($file,
+                                $lock_type,
+                                $blocking_timeout,
+                                $stale_lock_timeout,
+                                );
+
+Additionally, parameters may be passed as a hashref:
+  
+  my $lock = File::NFSLock->new({
+    file               => $file,
+    lock_type          => $lock_type,
+    blocking_timeout   => $blocking_timeout,
+    stale_lock_timeout => $stale_lock_timeout,
+  });
+
+=head1 PARAMETERS
 
 =over 4
 
-=item Parameter 1: filename
+=item Parameter 1: file
 
 Filename of the file upon which it is anticipated that a write will
 happen to.  Locking will provide the most recent version (uncached)
-of this file upon a successful file lock.
+of this file upon a successful file lock.  It is not necessary
+for this file to exist.
 
-=item Parameter 2: lock type
+=item Parameter 2: lock_type
 
 Lock type must be one of the following:
  
   BLOCKING
   BL
+  EXCLUSIVE (BLOCKING)
+  EX
   NONBLOCKING
   NB
   SHARED
   SH
+  Fcntl::LOCK_EX() (BLOCKING)
+  Fcntl::LOCK_NB() (NONBLOCKING)
+  Fcntl::LOCK_SH() (SHARED)
 
 Lock type determines whether the lock will be blocking, non blocking,
 or shared.  Blocking locks will wait until other locks are removed
@@ -243,13 +336,33 @@ another process currently has the lock.  Shared will allow other
 process to do a shared lock at the same time (shared is not yet
 implemented).
 
-=item Parameter 3: timeout (option)
+=item Parameter 3: blocking_timeout (optional)
 
 Timeout is used in conjunction with a blocking timeout.  If specified,
 File::NFSLock will block up to the number of seconds specified in
 timeout before returning undef (could not get a lock).
 
 
+=item Parameter 4: stale_lock_timeout (optional)
+
+Timeout is used to see if an existing lock file is older than the stale
+lock timeout.  If do_lock fails to get a lock, the modified time is checked
+and do_lock is attempted again.  If the stale_lock_timeout is set to low, a
+recursion load could exist so do_lock will only recurse 10 times (this is only
+a problem if the stale_lock_timeout is set too low -- on the order of one or two
+seconds).
+
+=head1 FAILURE
+
+On failure, a global variable, $File::NFSLock::errstr, should be set and should
+contain the cause for the failure to get a lock.  Useful primarily for debugging.
+
+=head1 LOCK_EXTENSION
+
+By default File::NFSLock will use a lock file extenstion of ".NFSLock".  This is
+now in a global variable $File::NFSLock::LOCK_EXTENSION that may be changed to
+suit other purposes (such as compatibility in mail systems).
+
 =head1 TODO
 
 Features yet to be implemented...
@@ -261,15 +374,6 @@ Features yet to be implemented...
 Need to allow for shared locking.  This will allow for safe
 reading on files.  Underway.
 
-=item Fnctl constants
-
-Allow for passing of Fnctl constants rather than keywords.
-
-=item Stale lock checking
-
-Allow for easy view into whether a lock is stale or not.  Stale
-locks can occur if process is "kill -9"ed during a lock.
-
 =item Tests
 
 Improve the test suite.
